/*
 *
 *    ******************************************************************************
 *    *
 *    * This program and the accompanying materials are made available under the
 *    * terms of the Apache License, Version 2.0 which is available at
 *    * https://www.apache.org/licenses/LICENSE-2.0.
 *    *
 *    *  See the NOTICE file distributed with this work for additional
 *    *  information regarding copyright ownership.
 *    * Unless required by applicable law or agreed to in writing, software
 *    * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *    * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *    * License for the specific language governing permissions and limitations
 *    * under the License.
 *    *
 *    * SPDX-License-Identifier: Apache-2.0
 *    *****************************************************************************
 *
 */

package net.brutex.gan;

import com.google.inject.Singleton;
import lombok.extern.slf4j.Slf4j;
import org.nd4j.linalg.api.ndarray.INDArray;

import javax.imageio.ImageIO;
import javax.swing.*;
import java.awt.*;
import java.awt.color.ColorSpace;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.UUID;

import static net.brutex.gan.App2.OUTPUT_DIR;
import static net.brutex.gan.App2Config.*;
@Slf4j
@Singleton
public class App2Display {

    private final JFrame frame = new JFrame();
  private final App2GUI display = new App2GUI();

  private final JPanel real_panel;
  private final JPanel fake_panel;


    public App2Display() {
        frame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
        frame.setContentPane(display.getOverall_panel());
        frame.setMinimumSize(new Dimension(300, 20));
        frame.pack();
        frame.setVisible(true);
        real_panel = display.getReal_panel();
        fake_panel = display.getGen_panel();
        real_panel.setLayout(new GridLayout(4, 4, 8, 8));
        fake_panel.setLayout(new GridLayout(4, 4, 8, 8));
    }

    public void visualize(INDArray[] samples, int batchElements, boolean isOrig) {
        for (INDArray sample : samples) {
            for(int i = 0; i<batchElements; i++) {
                final Image img = this.getImage(sample, i, isOrig);
                final ImageIcon icon = new ImageIcon(img);
                if(isOrig) {
                    if(real_panel.getComponents().length>=OUTPUT_PER_PANEL) {
                        real_panel.remove(0);
                    }
                    real_panel.add(new JLabel(icon));
                } else {
                        if(fake_panel.getComponents().length>=OUTPUT_PER_PANEL) {
                            fake_panel.remove(0);
                        }
                    fake_panel.add(new JLabel(icon));
                }
            }
        }
        frame.pack();
        frame.repaint();
    }

    public Image getImage(INDArray tensor, int batchElement, boolean isOrig) {
        final BufferedImage bi;
        if(CHANNELS >1) {
            bi = new BufferedImage(X_DIM, Y_DIM, BufferedImage.TYPE_INT_RGB); //need to change here based on channels
        } else {
            bi = new BufferedImage(X_DIM, Y_DIM, BufferedImage.TYPE_BYTE_GRAY); //need to change here based on channels
        }
        final int imageSize = X_DIM * Y_DIM;
        final int offset = batchElement * imageSize;
        int pxl = offset * CHANNELS; //where to start in the INDArray

        //Image in NCHW - channels first format
        for (int c = 0; c < CHANNELS; c++) { //step through the num channels for each pixel
            for (int y = 0; y < X_DIM; y++) { // step through the columns x
                for (int x = 0; x < Y_DIM; x++) { //step through the rows y
                    float f_pxl = tensor.getFloat(pxl) * COLORSPACE;
                    if(isOrig) log.trace("'{}.'{} Image (x,y,c): ({}, {}, {}) with INDArray with index {} and value '{}'", batchElement, isOrig ? "Real" : "Fake", x, y, c, pxl, f_pxl);
                    bi.getRaster().setSample(x, y, c, f_pxl);
                    pxl++; //next item in INDArray
                }
            }
        }
        ImageIcon orig = new ImageIcon(bi);
        Image imageScaled = orig.getImage().getScaledInstance((4 * X_DIM), (4 * Y_DIM), Image.SCALE_DEFAULT);
        ImageIcon scaled = new ImageIcon(imageScaled);
        //if(! isOrig)  saveImage(imageScaled, batchElement, isOrig);
        return imageScaled;

    }

    private static void saveImage(Image image, int batchElement, boolean isOrig) {
        String outputDirectory = OUTPUT_DIR; // Set the output directory where the images will be saved

        try {
            // Save the images to disk
            saveImage(image, outputDirectory, UUID.randomUUID().toString()+".png");

            log.debug("Images saved successfully.");
        } catch (IOException e) {
            log.error("Error saving the images: {}", e.getMessage());
        }
    }
    private static void saveImage(Image image, String outputDirectory, String fileName) throws IOException {
        File directory = new File(outputDirectory);
        if (!directory.exists()) {
            directory.mkdir();
        }

        File outputFile = new File(directory, fileName);
        ImageIO.write(imageToBufferedImage(image), "png", outputFile);
    }

    public static BufferedImage imageToBufferedImage(Image image) {
        if (image instanceof BufferedImage) {
            return (BufferedImage) image;
        }

        // Create a buffered image with the same dimensions and transparency as the original image
        BufferedImage bufferedImage;
        if (CHANNELS > 1) {
            bufferedImage =
                    new BufferedImage(
                            image.getWidth(null), image.getHeight(null), BufferedImage.TYPE_INT_ARGB);
        } else {
            bufferedImage =
                    new BufferedImage(
                            image.getWidth(null), image.getHeight(null), BufferedImage.TYPE_BYTE_GRAY);
        }

        // Draw the original image onto the buffered image
        Graphics2D g2d = bufferedImage.createGraphics();
        g2d.drawImage(image, 0, 0, null);
        g2d.dispose();

        return bufferedImage;
    }
}
